.dataset-selector.svelte-10ktfub{width:34rem}.question-input .bx--text-input{padding-left:2.75rem}.markdown.svelte-10ktfub{padding-top:1.5px;padding-bottom:1.5px}.markdown pre{overflow-x:auto;--tw-bg-opacity:1;background-color:rgb(226 232 240 / var(--tw-bg-opacity));padding:.5rem;font-size:.875rem;line-height:1.25rem;margin-top:.75rem;margin-bottom:.75rem}.markdown p,.markdown h1{background-color:inherit}.markdown p{margin-top:.75rem;font-size:.875rem;line-height:1.25rem;font-weight:inherit}.markdown ul{margin-top:.75rem;list-style-position:inside;list-style-type:disc}.markdown:has(+ .highlighted) p:last-child{display:inline!important}.highlighted+.markdown p:first-child{display:inline!important}.highlighted p{display:inline!important}
